""" Pydantic and Sqlalchemy models for the database

2. Location objects, where the forecast is for
8. GSP yield for storing GSP yield data

"""

import logging
from datetime import datetime
from typing import ClassVar, List, Optional

from pydantic import Field, field_validator
from sqlalchemy import Column, DateTime, Float, ForeignKey, Index, Integer, String
from sqlalchemy.orm import relationship

from nowcasting_datamodel.models.base import Base_Forecast
from nowcasting_datamodel.models.utils import CreatedMixin, EnhancedBaseModel
from nowcasting_datamodel.utils import datetime_with_timezone

logger = logging.getLogger(__name__)

########
# 2. Location
########


class LocationSQL(Base_Forecast):
    """Location that the forecast is for"""

    __tablename__ = "location"

    id = Column(Integer, primary_key=True)
    label = Column(String)
    gsp_id = Column(Integer, index=True, unique=True, nullable=False)
    gsp_name = Column(String, nullable=True)
    gsp_group = Column(String, nullable=True)
    region_name = Column(String, nullable=True)
    installed_capacity_mw = Column(Float, nullable=True)

    forecast = relationship("ForecastSQL", back_populates="location")
    gsp_yields = relationship("GSPYieldSQL", back_populates="location")
    metric_value = relationship("MetricValueSQL", back_populates="location")


class Location(EnhancedBaseModel):
    """Location that the forecast is for"""

    label: str = Field(..., description="")
    gsp_id: Optional[int] = Field(None, description="The Grid Supply Point (GSP) id", index=True)
    gsp_name: Optional[str] = Field(None, description="The GSP name")
    gsp_group: Optional[str] = Field(None, description="The GSP group name")
    region_name: Optional[str] = Field(None, description="The GSP region name")
    installed_capacity_mw: Optional[float] = Field(
        None, description="The installed capacity of the GSP in MW"
    )

    rm_mode: ClassVar[bool] = True

    def to_orm(self) -> LocationSQL:
        """Change model to LocationSQL"""

        if isinstance(self, LocationSQL):
            return self

        return LocationSQL(
            label=self.label,
            gsp_id=self.gsp_id,
            gsp_name=self.gsp_name,
            gsp_group=self.gsp_group,
            region_name=self.region_name,
            installed_capacity_mw=self.installed_capacity_mw,
        )


class GSPYieldSQL(Base_Forecast, CreatedMixin):
    """GSP Yield data"""

    __tablename__ = "gsp_yield"

    id = Column(Integer, primary_key=True)
    datetime_utc = Column(DateTime, index=True)
    solar_generation_kw = Column(Float)
    regime = Column(String, nullable=True)
    capacity_mwp = Column(Float, nullable=True)
    pvlive_updated_utc = Column(DateTime, nullable=True)

    # many (gsp_yields) to one (location)
    location = relationship("LocationSQL", back_populates="gsp_yields")
    location_id = Column(Integer, ForeignKey("location.id"), index=True)

    Index("ix_gsp_yield_datetime_utc_desc", datetime_utc.desc())


Index("ix_gsp_yield_created_utc", GSPYieldSQL.created_utc.desc())


class GSPYield(EnhancedBaseModel):
    """GSP Yield data"""

    datetime_utc: datetime = Field(..., description="The timestamp of the gsp yield")
    solar_generation_kw: float = Field(..., description="The amount of solar generation")
    regime: str = Field(
        "in-day", description="When the GSP data is pulled, can be 'in-day' or 'day-after'"
    )
    capacity_mwp: Optional[float] = Field(None, description="The estimate current capacity")
    pvlive_updated_utc: Optional[datetime] = Field(None, description="When PVlive made this value")
    gsp: Optional[Location] = Field(
        None,
        description="The GSP associated with this model",
    )

    @field_validator("datetime_utc", mode="before")
    def normalize_datetime_utc(cls, v):
        """Normalize datetime_utc field"""
        return datetime_with_timezone(cls, v)

    @field_validator("pvlive_updated_utc", mode="before")
    def normalize_pvlive_updated_utc(cls, v):
        """Normalize pvlive_updated_utc field"""
        return datetime_with_timezone(cls, v)

    @field_validator("solar_generation_kw")
    def validate_solar_generation_kw(cls, v):
        """Validate the solar_generation_kw field"""
        if v < 0:
            logger.debug(f"Changing solar_generation_kw ({v}) to 0")
            v = 0
        return v

    @field_validator("regime")
    def validate_regime(cls, v):
        """Validate the solar_generation_kw field"""
        if v not in ["day-after", "in-day"]:
            message = f"Regime ({v}) not in 'day-after', 'in-day'"
            logger.debug(message)
            raise Exception(message)
        return v

    def to_orm(self) -> GSPYieldSQL:
        """Change model to GSPYieldSQL"""
        return GSPYieldSQL(
            datetime_utc=self.datetime_utc,
            solar_generation_kw=self.solar_generation_kw,
            regime=self.regime,
            capacity_mwp=self.capacity_mwp,
            pvlive_updated_utc=self.pvlive_updated_utc,
        )


class LocationWithGSPYields(Location):
    """Location object with GSPYields"""

    gsp_yields: Optional[List[GSPYield]] = Field([], description="List of gsp yields")



""" Pydantic and Sqlalchemy models for the database

The following class are made
1. Reusable classes (utils.py)
2. Location objects, where the forecast is for (models.py)
3. Model object, what forecast model is used (models.py)
4. ForecastValue objects, specific values of a forecast and time (models.py)
5. Input data status, shows when the data was collected (models.py)
6. Forecasts, a forecast that is made for one gsp,
    for several time steps into the future (models.py)
7. PV system for storing PV data (pv.py)
8. PV yield for storing PV data (pv.py)

Current these models have a primary index of 'id'.
This keeps things very simple at the start.
But there can be multiple forecasts for similar values.

Later on it would be good to add a forecast latest table,
    where the latest forecast can be read.
The primary keys could be 'gsp_id' and 'target_datetime_utc'.
"""

from .api import *  # noqa F403
from .forecast import *  # noqa F403
from .gsp import *  # noqa F403
from .metric import *  # noqa F403
from .models import *  # noqa F403
from .pv import *  # noqa F403



""" Methods for adding adjust values to the forecast"""

import logging
from datetime import datetime, timedelta
from typing import List, Optional, Union

import numpy as np
import pandas as pd

from nowcasting_datamodel.models import Forecast, ForecastSQL, MetricValue, MetricValueSQL
from nowcasting_datamodel.read.read_metric import read_latest_me_national

logger = logging.getLogger()

MAX_ADJUST_PER = 0.2


def add_adjust_to_forecasts(
    forecasts_sql: List[ForecastSQL],
    session,
    max_adjust_percentage: Optional[float] = MAX_ADJUST_PER,
):
    """
    Adjust National Forecast by ME over the last week

    Note that no objects are returned as sqlalchemy objects can be adjust on the spot

    :param forecasts_sql: list of forecasts
    :param session: database sessions
    :param max_adjust_percentage: maximum percentage of forecast value that can be adjusted.
        If this is None, then no limit is used
    """
    logger.debug("Adding Adjusts to National forecast")

    # get the national forecast only
    forecast_national = [f for f in forecasts_sql if f.location.gsp_id == 0]

    if len(forecast_national) != 1:
        logger.debug("Could not find single national forecast, tehre fore not adding adjust")

    add_adjust_to_national_forecast(
        forecast=forecast_national[0], session=session, max_adjust_percentage=max_adjust_percentage
    )


def add_adjust_to_national_forecast(
    forecast: ForecastSQL, session, max_adjust_percentage: Optional[float] = MAX_ADJUST_PER
):
    """
    Add adjust to national forecast.

    1. Get latest me results for time_of_day and forecast horizont
    2. Reduce latest me results relevant for now
    3. Reduce adjust value by at most X% of forecast_value.expected_power_generation_megawatts,
    4. Add Me to adjust_mw property in ForecastValueSQL

    :param forecast: national forecast
    :param session:
    :param max_adjust_percentage: maximum percentage of forecast value that can be adjusted.
        If this is None, then no limit is used

    :return:
    """

    # get the target time and model name
    datetime_now = forecast.forecast_values[0].target_time
    last_datetime = forecast.forecast_values[-1].target_time
    model_name = forecast.model.name

    logger.debug(
        f"Adding adjuster to national forecast for "
        f"{datetime_now} and {model_name}, {last_datetime=}"
    )

    # 1. read metric values
    latest_me = read_latest_me_national(session=session, model_name=model_name)
    if len(latest_me) == 0:
        logger.warning(f"Found no ME values found for {model_name=}")
    else:
        logger.debug(f"Found {len(latest_me)} latest ME values")

    # 2. filter value down to now onwards
    # get the number of hours to go ahead, we've added 1 to make sure we use the last one as well
    hours_ahead = get_forecast_horizon_from_forecast(forecast)
    # change to dataframe
    latest_me_df = reduce_metric_values_to_correct_forecast_horizon(
        latest_me=latest_me, datetime_now=datetime_now, hours_ahead=hours_ahead
    )
    assert len(latest_me_df) > 0

    # 3. add values to forecast_values
    for forecast_value in forecast.forecast_values:
        target_time = forecast_value.target_time
        try:
            # get value from df
            value = latest_me_df[latest_me_df["datetime"] == forecast_value.target_time]
            value = value.iloc[0].value

            # reduce adjust value by at most 20% of expected_power_generation_megawatts,
            # if the forecast_value is not 0
            # note that adjust value can be negative, so we need to be careful
            # also, if the forecast value is 0, then adjust value should also be 0
            if max_adjust_percentage is not None:
                if forecast_value.expected_power_generation_megawatts != 0:
                    max_adjust = (
                        max_adjust_percentage * forecast_value.expected_power_generation_megawatts
                    )
                    if value > max_adjust:
                        value = max_adjust
                    elif value < -max_adjust:
                        value = -max_adjust
                else:
                    value = 0.0

            # add value to ForecastValueSQL
            forecast_value.adjust_mw = float(value)
            if np.isnan(value):
                logger.debug(
                    f"Found ME value for {target_time} in {latest_me_df}, "
                    f"but it was NaN, therefore adding adjust_mw as 0"
                )
                forecast_value.adjust_mw = 0.0

        except Exception:
            logger.debug(
                f"Could not find ME value for {target_time} in {latest_me_df}, "
                f"therefore adding adjust_mw as 0"
            )
            forecast_value.adjust_mw = 0.0


def get_forecast_horizon_from_forecast(forecast: Union[ForecastSQL, Forecast]):
    """
    Get the number of hours ahead from the forecast

    :param forecast:
    :return:
    """
    datetime_now = forecast.forecast_values[0].target_time
    last_datetime = forecast.forecast_values[-1].target_time
    hours_ahead = int((last_datetime - datetime_now).seconds / 3600) + 1
    hours_ahead += (last_datetime - datetime_now).days * 24

    logger.debug(f"Hours ahead is {hours_ahead}, {last_datetime=}, {datetime_now=} ")

    return hours_ahead


def reduce_metric_values_to_correct_forecast_horizon(
    latest_me: List[MetricValueSQL], datetime_now: datetime, hours_ahead=4
) -> pd.DataFrame:
    """
    Change all latest ME results 2D to 1D array

    Start:
    # time forecast_horizon value
    # 00.00 0  0.1
    # 00.00 30 0.2
    # 00.00 60 0.6
    # .....
    # 00.30 0  0.11
    # 00.30 30 0.21
    # 00.30 60 0.61
    # .....
    # 23.30 0  0.11
    # 23.30 30 0.21
    # 23.30 60 0.61

    Using datetime_now of 04:30

    End:
    # time forecast_horizon value
    # 04.30 0  0.1
    # 05.00 30 0.2
    # 05.30 60 0.6
    # .....

    :param latest_me: list of latest me results for all time_of_day and all forecast_horizons
    :param datetime_now: the dateteim now, so we know where to use forecast_horizons of 0
    :param hours_ahead: how many hours ahead
    :return: dataframe containing
        'datetime'
        'time_of_day'
        'forecast_horizon'
        'value'
    """

    logger.debug(
        f"Reducing metric values to correct forecast horizon {datetime_now=} {hours_ahead=}"
    )

    latest_me_df = pd.DataFrame(
        [MetricValue.model_validate(m, from_attributes=True).model_dump() for m in latest_me]
    )
    if len(latest_me_df) == 0:
        # no latest ME values, so just making an empty dataframe
        latest_me_df = pd.DataFrame(columns=["forecast_horizon_minutes", "time_of_day", "value"])

    # Let now big a dataframe of datetimes from now onwards. Lets say the time is 04.30, then
    # time forecast_horizon value
    # 04.30 0  0.1
    # 05.00 30 0.2
    # 05.30 60 0.6
    # .....

    results_df = pd.DataFrame(
        index=pd.date_range(
            start=datetime_now, end=datetime_now + timedelta(hours=hours_ahead), freq="30T"
        ),
    )
    results_df["datetime"] = results_df.index
    results_df["time_of_day"] = results_df.index.time
    results_df["forecast_horizon_minutes"] = range(0, 30 * len(results_df), 30)
    latest_me_df["forecast_horizon_minutes"] = latest_me_df["forecast_horizon_minutes"].astype(int)
    results_df["forecast_horizon_minutes"] = results_df["forecast_horizon_minutes"].astype(int)
    latest_me_df["value"] = latest_me_df["value"].astype(float)

    all_df = pd.merge(
        results_df,
        latest_me_df,
        how="left",
        on=["time_of_day", "forecast_horizon_minutes"],
    )

    all_df = all_df[["datetime", "time_of_day", "value", "forecast_horizon_minutes"]]

    return all_df



""" Read pv functions """

import logging
from datetime import datetime, timezone
from typing import List, Optional, Union

import pandas as pd
from sqlalchemy import desc, func
from sqlalchemy.orm import Session, contains_eager, joinedload

from nowcasting_datamodel.models import GSPYield, GSPYieldSQL, LocationSQL

logger = logging.getLogger(__name__)


def get_latest_gsp_yield(
    session: Session,
    gsps: Union[List[LocationSQL], List[int]],
    append_to_gsps: bool = False,
    regime: str = "in-day",
    datetime_utc: Optional[datetime] = None,
    allow_nans_in_capacities: bool = True,
) -> Union[List[GSPYieldSQL], List[LocationSQL]]:
    """
    Get the last gsp yield data

    :param session: database sessions
    :param gsps: list of gsps
    :param append_to_gsps: append gsp yield to pv systems, or return pv systems.
        If appended the yield is access by 'pv_system.last_gsp_yield'
    :param regime: What regime the data is in, either 'in-day' or 'day-after'
    :param datetime_utc: Optional to filter on datetime
    :param allow_nans_in_capacities: Optional to allow nans in capacities. Default is True
    :return: either list of gsp yields, or pv systems
    """

    if isinstance(gsps[0], int):
        gsp_ids = gsps
    else:
        gsp_ids = [gsp.gsp_id for gsp in gsps]

    # start main query
    query = session.query(GSPYieldSQL)
    query = query.join(LocationSQL)
    query = query.where(
        LocationSQL.id == GSPYieldSQL.location_id,
    )

    # filter on regime
    query = query.where(GSPYieldSQL.regime == regime)

    # only select on results per pv system
    query = query.distinct(LocationSQL.gsp_id)

    # select only th epv systems we want
    query = query.where(LocationSQL.gsp_id.in_(gsp_ids))

    if datetime_utc is not None:
        # filter on datetime
        query = query.where(GSPYieldSQL.datetime_utc >= datetime_utc)

    # don't get any nans. (Note nan+1 > nan = False)
    if not allow_nans_in_capacities:
        query = query.where(GSPYieldSQL.capacity_mwp + 1 > GSPYieldSQL.capacity_mwp)

    # order by 'created_utc' desc, so we get the latest one
    query = query.order_by(
        LocationSQL.gsp_id,
        desc(GSPYieldSQL.datetime_utc),
        desc(GSPYieldSQL.created_utc),
    )

    # get all results
    gsp_yields: List[GSPYieldSQL] = query.all()

    logger.debug(f"Found {len(gsp_yields)}  latest gsp yields")

    # add utc timezone
    for gsp_yield in gsp_yields:
        gsp_yield.datetime_utc = gsp_yield.datetime_utc.replace(tzinfo=timezone.utc)

    if not append_to_gsps:
        return gsp_yields
    else:
        # get list of gsp with last pv yields
        gsp_systems_with_gsp_yields = []
        for gsp_yield in gsp_yields:
            gsp = gsp_yield.location
            gsp.last_gsp_yield = gsp_yield

            gsp_systems_with_gsp_yields.append(gsp)

        logger.debug(f"Found {len(gsp_systems_with_gsp_yields)} gsps with yields")

        # add pv systems that dont have any pv yields
        gsp_ids_with_gsp_yields = [gsp.gsp_id for gsp in gsp_systems_with_gsp_yields]

        gsp_systems_with_no_gsp_yields = []
        for gsp in gsps:
            if gsp.gsp_id not in gsp_ids_with_gsp_yields:
                gsp.last_gsp_yield = None

                gsp_systems_with_no_gsp_yields.append(gsp)

        logger.debug(f"Found {len(gsp_systems_with_no_gsp_yields)} gsps with no yields")

        all_gsp_systems = gsp_systems_with_gsp_yields + gsp_systems_with_no_gsp_yields

        return all_gsp_systems


def get_gsp_yield(
    session: Session,
    gsp_ids: List[int],
    start_datetime_utc: datetime,
    regime: Optional[str] = None,
    end_datetime_utc: Optional[datetime] = None,
    filter_nans: Optional[bool] = True,
) -> List[GSPYieldSQL]:
    """
    Get the gsp yield values.

    :param session: sqlalmcy sessions
    :param gsp_ids: list of gsp ids that we filter on
    :param start_datetime_utc: filter values on this start datetime
    :param regime: filter query on this regim. Can be "in-day" or "day-after"
    :param end_datetime_utc: optional end datetime filter
    :param filter_nans: optional filter out nans. Default is True
    :return: list of gsp yields
    """

    # start main query
    query = session.query(GSPYieldSQL)
    query = query.join(LocationSQL)
    query = query.options(joinedload(GSPYieldSQL.location))
    query = query.where(
        LocationSQL.id == GSPYieldSQL.location_id,
    )

    # filter on regime
    if regime is not None:
        query = query.where(GSPYieldSQL.regime == regime)

    # filter on datetime
    query = query.where(GSPYieldSQL.datetime_utc >= start_datetime_utc)
    if end_datetime_utc is not None:
        query = query.where(GSPYieldSQL.datetime_utc <= end_datetime_utc)

    # don't get any nans. (Note nan+1 > nan = False)
    if filter_nans:
        query = query.where(GSPYieldSQL.solar_generation_kw + 1 > GSPYieldSQL.solar_generation_kw)

    # only select on results per gsp
    query = query.distinct(*[LocationSQL.gsp_id, GSPYieldSQL.datetime_utc])

    # select only the gsp systems we want
    query = query.where(LocationSQL.gsp_id.in_(gsp_ids))

    # order by 'created_utc' desc, so we get the latest one
    query = query.order_by(
        LocationSQL.gsp_id,
        desc(GSPYieldSQL.datetime_utc),
        desc(GSPYieldSQL.created_utc),
    )

    # get all results
    gsp_yields: List[GSPYieldSQL] = query.all()

    for gsp_yield in gsp_yields:
        gsp_yield.datetime_utc = gsp_yield.datetime_utc.replace(tzinfo=timezone.utc)

    return gsp_yields


def get_gsp_yield_by_location(
    session: Session,
    gsp_ids: List[int],
    start_datetime_utc: datetime,
    regime: Optional[str] = None,
    end_datetime_utc: Optional[datetime] = None,
    filter_nans: Optional[bool] = True,
) -> List[LocationSQL]:
    """
    Get the locations with gsp yield values.

    :param session: sqlalmcy sessions
    :param gsp_ids: list of gsp ids that we filter on
    :param start_datetime_utc: filter values on this start datetime
    :param regime: filter query on this regim. Can be "in-day" or "day-after"
    :param end_datetime_utc: optional end datetime filter
    :param filter_nans: optional filter out nans. Default is True
    :return: list of locations with gsp yields
    """

    # start main query
    query = session.query(LocationSQL)
    query = query.join(GSPYieldSQL)
    query = query.where(LocationSQL.id == GSPYieldSQL.location_id)

    # filter on regime
    if regime is not None:
        query = query.where(GSPYieldSQL.regime == regime)

    # filter on datetime
    query = query.where(GSPYieldSQL.datetime_utc >= start_datetime_utc)
    if end_datetime_utc is not None:
        query = query.where(GSPYieldSQL.datetime_utc <= end_datetime_utc)

    # don't get any nans. (Note nan+1 > nan = False)
    if filter_nans:
        query = query.where(GSPYieldSQL.solar_generation_kw + 1 > GSPYieldSQL.solar_generation_kw)

    # only select on results per gsp
    query = query.distinct(*[LocationSQL.gsp_id, GSPYieldSQL.datetime_utc])

    # select only the gsp systems we want
    query = query.where(LocationSQL.gsp_id.in_(gsp_ids))

    # load only once
    query = query.options(contains_eager(LocationSQL.gsp_yields))

    # order by 'created_utc' desc, so we get the latest one
    query = query.order_by(
        LocationSQL.gsp_id,
        desc(GSPYieldSQL.datetime_utc),
        desc(GSPYieldSQL.created_utc),
    )

    # get all results
    locations: List[LocationSQL] = query.all()

    for location in locations:
        for gsp_yield in location.gsp_yields:
            gsp_yield.datetime_utc = gsp_yield.datetime_utc.replace(tzinfo=timezone.utc)

    return locations


def get_gsp_yield_sum(
    session: Session,
    gsp_ids: List[int],
    start_datetime_utc: datetime,
    regime: Optional[str] = None,
    end_datetime_utc: Optional[datetime] = None,
) -> List[GSPYield]:
    """
    Get the sum of gsp yield values.

    :param session: sqlalchemy sessions
    :param gsp_ids: list of gsp ids that we filter on
    :param start_datetime_utc: filter values on this start datetime
    :param regime: filter query on this regim. Can be "in-day" or "day-after"
    :param end_datetime_utc: optional end datetime filter

    :return: list of GSPYield objects
    """

    logger.info(f"Getting gsp yield sum for {len(gsp_ids)} gsp systems")

    if regime is None:
        logger.debug("No regime given, defaulting to 'in-day'")
        regime = "in-day"

    # start main query
    query = session.query(
        GSPYieldSQL.datetime_utc,
        func.sum(GSPYieldSQL.solar_generation_kw).label("solar_generation_kw"),
    )

    # join with location table
    query = query.join(LocationSQL)

    # select only the gsp systems we want
    query = query.where(LocationSQL.gsp_id.in_(gsp_ids))

    # filter on regime
    query = query.where(GSPYieldSQL.regime == regime)

    # filter on datetime
    query = query.where(GSPYieldSQL.datetime_utc >= start_datetime_utc)
    if end_datetime_utc is not None:
        query = query.where(GSPYieldSQL.datetime_utc <= end_datetime_utc)

    # filter out nans
    query = query.where(GSPYieldSQL.solar_generation_kw + 1 > GSPYieldSQL.solar_generation_kw)

    # group and order by datetime
    query = query.group_by(GSPYieldSQL.datetime_utc)
    query = query.order_by(GSPYieldSQL.datetime_utc)

    results = query.all()

    # format results
    results = [
        GSPYield(
            datetime_utc=result.datetime_utc,
            solar_generation_kw=result.solar_generation_kw,
            regime=regime,
        )
        for result in results
    ]

    return results


def get_latest_gsp_capacities(
    session: Session, gsp_ids: List[int], datetime_utc: Optional[datetime] = None
) -> pd.Series:
    """
    Get the latest gsp capacities.

    :param session: database sessions
    :param gsp_ids: list of gsp_ids
    :param datetime_utc: datetime to filter on
    :return: pd.Series with gsp capacities, gsp_ids as the index
    """

    gsp_yields = get_latest_gsp_yield(
        session=session, gsps=gsp_ids, datetime_utc=datetime_utc, allow_nans_in_capacities=False
    )

    # format results
    eff_capacities = [gsp_yield.capacity_mwp for gsp_yield in gsp_yields]
    gsp_ids_from_db = [gsp_yield.location.gsp_id for gsp_yield in gsp_yields]

    return pd.Series(eff_capacities, index=gsp_ids_from_db)



from datetime import datetime, timezone

from nowcasting_datamodel.fake import (
    generate_fake_forecasts,
    make_fake_forecast,
    make_fake_forecast_value,
    make_fake_forecasts,
    make_fake_gsp_yields,
    make_fake_input_data_last_updated,
    make_fake_intensity,
    make_fake_location,
    make_fake_national_forecast,
)
from nowcasting_datamodel.models import (
    Forecast,
    ForecastSQL,
    ForecastValue,
    ForecastValueLatestSQL,
    GSPYieldSQL,
    InputDataLastUpdated,
    InputDataLastUpdatedSQL,
    Location,
    LocationSQL,
)
from nowcasting_datamodel.models.forecast import ForecastValueSQL


def test_make_fake_intensity():
    assert make_fake_intensity(datetime(2022, 1, 1)) == 0
    assert make_fake_intensity(datetime(2022, 1, 1, 6)) == 0
    assert 0 < make_fake_intensity(datetime(2022, 1, 1, 9)) < 1
    assert make_fake_intensity(datetime(2022, 1, 1, 12)) == 1
    assert 0 < make_fake_intensity(datetime(2022, 1, 1, 15)) < 1
    assert make_fake_intensity(datetime(2022, 1, 1, 18)) == 0
    assert make_fake_intensity(datetime(2022, 1, 2)) == 0


def test_make_fake_location():
    location_sql: LocationSQL = make_fake_location(1)
    location = Location.model_validate(location_sql, from_attributes=True)
    _ = Location.to_orm(location)


def test_make_fake_input_data_last_updated():
    input_sql: InputDataLastUpdatedSQL = make_fake_input_data_last_updated()
    input = InputDataLastUpdated.model_validate(input_sql, from_attributes=True)
    _ = InputDataLastUpdated.to_orm(input)


def test_make_fake_forecast_value():
    target = datetime(2023, 1, 1, tzinfo=timezone.utc)

    forecast_value_sql: ForecastValueSQL = make_fake_forecast_value(target_time=target)
    forecast_value = ForecastValue.model_validate(forecast_value_sql, from_attributes=True)
    _ = ForecastValue.to_orm(forecast_value)


def test_make_fake_forecast(db_session):
    forecast_sql: ForecastSQL = make_fake_forecast(gsp_id=1, session=db_session)
    forecast = Forecast.model_validate(forecast_sql, from_attributes=True)
    forecast_sql = Forecast.to_orm(forecast)
    _ = Forecast.model_validate(forecast_sql, from_attributes=True)

    from sqlalchemy import text

    f = ForecastValueSQL(target_time="2023-01-01 12:00:00", expected_power_generation_megawatts=3)
    db_session.add(f)
    db_session.commit()
    _ = db_session.execute(text("SELECT * FROM forecast_value_2023_01")).all()


def test_make_fake_forecast_with_creation_time(db_session):
    t0 = datetime(2023, 1, 1, tzinfo=timezone.utc)
    forecast_creation_time = datetime(2023, 1, 1, 1, tzinfo=timezone.utc)
    forecast_sql: ForecastSQL = make_fake_forecast(gsp_id=1, session=db_session, t0_datetime_utc=t0)
    assert forecast_sql.forecast_creation_time == t0

    forecast_sql: ForecastSQL = make_fake_forecast(
        gsp_id=1,
        session=db_session,
        t0_datetime_utc=t0,
        forecast_creation_time=forecast_creation_time,
    )
    assert forecast_sql.forecast_creation_time == forecast_creation_time


def test_generate_fake_forecasts(db_session):
    fake_forecasts = generate_fake_forecasts(
        session=db_session, gsp_ids=[0, 1, 2, 3], add_latest=True
    )

    assert len(db_session.query(ForecastSQL).all()) == 0
    assert len(db_session.query(ForecastValueSQL).all()) == 0
    assert len(db_session.query(ForecastValueLatestSQL).all()) == 0
    assert len(fake_forecasts) == 4


def test_make_fake_forecasts(db_session):
    make_fake_forecasts(session=db_session, gsp_ids=[0, 1, 2, 3], add_latest=True)

    assert len(db_session.query(ForecastSQL).all()) == 4
    assert len(db_session.query(ForecastValueSQL).all()) == 4 * 112
    assert len(db_session.query(ForecastValueLatestSQL).all()) == 4 * 112


def test_make_fake_forecasts_without_latest(db_session):
    make_fake_forecasts(session=db_session, gsp_ids=[0, 1, 2, 3], add_latest=False)

    assert len(db_session.query(ForecastSQL).all()) == 4
    assert len(db_session.query(ForecastValueSQL).all()) == 4 * 112
    assert len(db_session.query(ForecastValueLatestSQL).all()) == 0


def test_make_national_fake_forecast(db_session):
    forecast_sql: ForecastSQL = make_fake_national_forecast(session=db_session)
    forecast = Forecast.model_validate(forecast_sql, from_attributes=True)
    _ = Forecast.to_orm(forecast)


def test_make_fake_gsp_yields(db_session):
    make_fake_gsp_yields(session=db_session, gsp_ids=list(range(0, 10)))

    assert len(db_session.query(LocationSQL).all()) == 10
    # 2 days for each location, both regimes
    assert len(db_session.query(GSPYieldSQL).all()) == 10 * 48 * 2 * 2



""" Usual setup file for package """

from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent
install_requires = (this_directory / "requirements.txt").read_text().splitlines()
long_description = (this_directory / "README.md").read_text()
#

with open("nowcasting_datamodel/__init__.py") as f:
    for line in f:
        if line.startswith("__version__"):
            _, _, version = line.replace("'", "").split()
            version = version.replace('"', "")


setup(
    name="nowcasting_datamodel",
    packages=find_packages(),
    version=version,
    license="MIT",
    description="Data Model for the OCF nowcasting project",
    author="Peter Dudfield",
    author_email="peter@openclimatefix.org",
    company="Open Climate Fix Ltd",
    url="https://github.com/openclimatefix/nowcasting_datamodel",
    keywords=[
        "SQL",
        "Datamodel",
    ],
    install_requires=install_requires,
    long_description=long_description,
    long_description_content_type="text/markdown",
)
